# Sample code for Hitachi Hokudai Lab. and Hokkaido University Contest 2020 A

* This is an explanation of the sample code used in the Hitachi Hokudai Lab. and Hokkaido University Contest 2020 A. Based on the distributed code the participants can create input samples and evaluate the score of their output on their own local machine.
* Please note that you are using the provided sample code on your own responsibility. In particular, neither Hokkaido University nor Hitachi Ltd. can be held responsible or liable, in case the execution of the provided sample code causes any damage to your computer.
* The provided program for score evaluation is identical with the program which will be used to evaluate your score during the contest. However, since the random seed used to generate the input cases during the contest may vary from your input cases, the final score of your program might vary.

## Sample code overview

The sample code is written in C++ and can be found in the subfolder `src`.

* `judge_A.cpp`
    * This program is the game engine which simulates the environment and scores the contestant output.
* `judge_A.lib/`
    * It contains the header files needed by `judge_A.cpp`.
* `generator_A.cpp`
    * A program that generates input samples. A random seed for generating the input sample and DayType can be set with optional arguments.
    Note: The input samples generated here are files for the judge to read and do not match the data format output by the judge to the contestant. For details, please refer to "Specifications of input samples" below.
* `generator_A.lib/`
    * It contains the header files needed by `generator_A.cpp`.
* `sample_A.cpp`
    * This is an example of the contestant code. It always outputs `stay` in response to inputs from judge. You can use the sample when submitting your code.
* `sample_A.py`
    * This is an example of the contestant code written in Python. It always outputs `stay` in response to inputs from judge. You can use the sample when submitting your code.

## Compilation

To create executable files `generator_A`, `judge_A`, and `sample_A`, please compile generator_A.cpp, judge_A.cpp, and sample_A.cpp as follows.

```bash
g++ -Isrc -std=c++17 -O2 src/generator_A.cpp -o generator_A
g++ -Isrc -std=c++17 -O2 src/judge_A.cpp -o judge_A
g++ -Isrc -std=c++17 -O2 src/sample_A.cpp -o sample_A
```

`make` also can be used.

```bash
make
```

If `make` is used, generator_A.cpp, judge_A.cpp, and sample_A.cpp will be compiled at the same time.
Separately, please create the contestant code to solve the problem and compile it, if necessary, to make an executable file.

Usage of the toolkit is explained in the following. 
Henceforth, the contestant executable file will be referred to as `sample_A`.

## Generation of input samples and interactive processing

Please generate an input sample as follows:

``` bash
./generator_A [options] > test_case.in
```

The available optional arguments are:

* `-s seed_number --seed=seed_number`
  * Specify a random seed for generating an input sample. If the seed is specified at the same time as `-r`,` -r` has priority. The default random seed is `23587163955511535`
* `-r --random`
  * Get a random seed from the system-dependent random number generator. If `-r` is specified at the same time as `-s`,` -r` has priority.
* `-d day_type, ex) --day-type=DayType (= 0, 1, 2, 3)`
  * Specify the DayType. If this optional argument is not specified, the DayType will be determined randomly.
* `-D, --deterministic`
  * The difference between generated and consumed powers will be always fixed to the predicted value. In other words, stochastic fluctuations will be always zero and downpour (or sudden clear sky) will not occur.
* `-? --help`
  * Display a help message.

For input sample `test_case.in`, the score is calculated as follows.

```bash
./judge_A ./sample_A < test_case.in > test_case.result
```

If the contestant code is written in Python etc., the score is calculated as follows. (Note that enclose the command in quotation marks.)

```bash
./judge_A 'python3 ./src/sample_A.py' < test_case.in > test_case.result
```

In both cases, the score is output to `test_case.result`.

### Specifications of input samples

The contents of input samples, which are generated by `generator_A`, are described in this section. 

First, `generator_A` outputs parameters included in "Input and Output format 1" on the web page "A - hokudai_hitachi2020_a"(https://atcoder.jp/contests/hokudai-hitachi2020/tasks/hokudai_hitachi2020_a). 
The content output here conforms to the format of problem B ("Input and Output format 1"). Regarding problem A, some unnecessary data is included.

Next, `generator_A` outputs the following information. 
For implementation reasons, the order information is output and the judge reads it, but never outputs it to the contestant.

1. Order information

  ```
  $N_{\rm order}^1$
  ${\rm id}^1_1 {\rm from}^1_1 {\rm to}^1_1 t$
  :
  ${\rm id}^1_{N_{\rm order}^1} {\rm from}^1_{N_{\rm order}^1} {\rm to}^1_{N_{\rm order}^1} t$
  :
  :
  $N_{\rm order}^{T_{\rm last}}$
  ${\rm id}^{T_{\rm last}}_1 {\rm from}^{T_{\rm last}}_1 {\rm to}^{T_{\rm last}}_1 t$
  :
  ${\rm id}^{T_{\rm last}}_{N_{\rm order}^1} {\rm from}^{T_{\rm last}}_{N_{\rm order}^1} {\rm to}^{T_{\rm last}}_{N_{\rm order}^1} t$
  ```

  * $N_{\rm order}^t$ is the number of orders that occur at time $t$.
  If it is $0$, the details of the orders will not be output.
  * The following $N_{\rm order}^t$ lines show the details of the orders that occurred at time $t$.
  For the $n$-th order that occurred at time $t$
    * ${\rm id}_n^t$ is the assigned id.
    * ${\rm from}_n^t$ is the starting point.
    * ${\rm to}_n^t$ is the arrival point.
    * $t$ is the time of occurrence.

2. Difference between generated and consumed powers in nanogrids

  ```
  $x_1$ ${\rm pw}^{{\rm actual}, 1}_1$ ... $x_{N_{\rm grid}}$ ${\rm pw}^{{\rm actual}, 1}_{N_{\rm grid}}$
  :
  $x_1$ ${\rm pw}^{{\rm actual}, T_{\rm last}}_1$ ... $x_{N_{\rm grid}}$ ${\rm pw}^{{\rm actual}, T_{\rm last}}_{N_{\rm grid}}$
  ```

  * In the $t$-th line, the actual difference between generated and consumed powers in each nanogrid at time $t$ is displayed. It is the sum of the predicted difference, a stochastic fluctuation ($\delta$), and the variation due to sudden downpour and unexpected sunny weather ($\pm \Delta_{\rm event}$, see the web page "A - hokudai_hitachi2020_a"(https://atcoder.jp/contests/hokudai-hitachi2020/tasks/hokudai_hitachi2020_a) for detail).
  * $x_i$ represents the location (vertex id) of the $i$-th nanogrid. ${\rm pw_i^{{\rm actual}, t}}$ represents the actual difference between generated and consumed powers in the $i$-th nanogrid at time $t$.

## About sample code

The sample_A.cpp is a sample of contestant code. 
It is designed so that you can focus on describing the operation of EV.
The sample code is written using C++17.

### Design

The information from the judge ("Input and Output format 1") is read by the constructor of the `A` class.
The `A` class internally has instances of the` graph_data` class, which holds information about a graph, the `grid_data` class, which holds information about nanogrids, and the` EV_data` class, which holds information about EVs.

The information from the judge for each turn ("Input and Ouptut Format 2") is read by three classes: the `grid_info` class that handles information about the nanogrid, the` EV_info` class that handles information about EV, and the `order_info' class that handles information about orders.

The `<command>` to each EV for each turn ("Input and Ouptut Format 2") is taken out from the front the `command_queue` and output to the judge.
The `command_queue` is held inside the `strategy` class.

The `<command>` is inserted into the `command_queue` using the `command` member function. Users can implement their own strategies by inheriting the `strategy` class and overriding the `command` member functions. The implemented class can actually be active by assigning an instance of the class to the `str` shared pointer variable.

Below is a list of functions and classes that are supposed to be important when implementing your own strategy. It doesn't cover all, so please read the source code for the details of functions and classes that are not explained.

#### Member function `strategy::command`

```c++
void strategy::command(const grid_info&, const EV_info& , const order_info&);
```

* It is a core function for implementing a strategy and is called for every turn. On the basis of the information from the judge for each turn, 
decide the `<command>` to push back to the the `command_queue`. 

#### Member function `strategy::enqueue`

```c++
void strategy::enqueue(size_t EV_index, const string& cmd);
```

* It is used for queuing `<command>`. `cmd` is inserted into the `command_queue` of the EV specified by `EV_index`.
* No line break is required at the end of `cmd`.
* Note that `EV_index` is 0-index.

```c++
void enqueue(size_t EV_index, const string& cmd, size_t repeat)
```

* It is used for queuing multiple consecutive `<command>`s. Multiple `cmd`s are inserted into the `command queue` of the EV specified by `EV_index`. The number of inserted `cmd`s is specified by `repeat`.
* It has been implemented since there are many consecutive `<command>`s, such as `move XXX`,` charge_from_grid XXX`, and `charge_to_grid XXX`.

```c++
void enqueue(size_t EV_index, list<string>&& cmd_list)
```

* It is used for queuing a list of `<command>`s. A list of `<command>`s (`cmd_list`) is inserted into the `command_queue` of the EV specified by `EV_index`.
* This function can be used for queuing more complicated processes at once. It is convenient to combine it with `move_to` described later.

#### Member function `bool strategy::is_free`

```c++
bool strategy::is_free(size_t EV_index);
```

* Is is used to check the status of `command_queue`. It returns `true` if 
`command_queue` of the EV specified by `EV_index` is empty, otherwise it returns `false`.
* You can check EVs that have no process currently in progress.

#### Class `graph_summary`

```c++
struct graph_summary {
    vector<vector<size_t>> len;
    vector<vector<size_t>> next;
    vector<size_t> nanogrid_pos;
    size_t diameter = 0;
    size_t cover_radius = 0;

    graph_summary(const graph_data& graph, const grid_data& grid);
};
```

* It holds summary information about the graph. 
* The Warshall–Floyd algorithm is executed in the constructor and it holds:
  * `len`: Shortest path length from vertex A to vertex B is stored in `len[A][B]`.
  * `next`: `next[A][B]` represents the next visiting vertex on the shortest path from vertex A to vertex B. Please refer to the implementation of `move_EV` for how to use it.
  * `nanogrid_pos`: The vertex id where each nanogrid exists.
  * `diameter`: It is the diameter of the graph. (The diameter is the maximum of the shortest path lengths between all pairs of vertices in the graph.)
  * `cover_radius`: It represents the longest of the shortest path lengths to the nearest nanogrid. In other words, When the EV moves from any vertex on the graph to the nearest nanogrid, its shortest path length does not exceed `cover_radius`.

#### Class `struct action`

``` c++
struct action : std::list<std::string> {};
```

* A placeholder to hide the `<command>` generation for complex EV operations.

#### Class `struct move_EV`

``` c++
struct move_EV : action {
    move_EV(size_t current, size_t goal, const graph_summary& gs);
    move_EV(size_t current, const std::vector<size_t>& path, const graph_summary& gs);
};
```

* A class to hide the `<command>` generation of complex EV moves.

```c++
move_EV::move_EV(size_t current, size_t goal, const graph_summary& gs);
```

* Generates a set of `<command>`s to travel on the shortest path from the `current` vertex to the` goal` vertex.

```c++
move_EV::move_EV(size_t current, const std::vector<size_t>& path, const graph_summary& gs);
```

* Generates a set of `<command>`s to travel on the shortest path from starting `current` vertex to the vertices specified by `path` in orders.

### Strategies implemented in the sample code

1. `all_stay`
    * It always outputs `stay` for all EVs.
    * It is a common strategy for problem A and problem B.
2. `random_walk`
    * Each EV charges the power needed to move for 50 time steps on the nearest nanogrid as it reaches a vertex.
    * The fully charged EV randomly determines the destination and sets it in the `command_queue`.
    * Each EV repeats the above actions.
    * It is a common strategy for problem A and problem B.
3. `transport_only_0`
    1. If the EV is not currently on the nanogrid, it will move to the nearest nanogrid.
    2. If the remaining power of the EV is below a threshold, the EV will be charged to exceed the threshold.
    3. If there are orders to which any EV is not assigned, the EV will carry up to 4 packages in order of their order ID. If the EVs are not charged enough to carry them, charge the EVs enough before starting.
    * The actual transportation route is determined by the greedy algorithm.
    * It's a strategy dedicated to problem B.


## Acknowledgement

* The `testlib.hpp` included in this toolkit is the modified version of the code published under the MIT license at `https://github.com/MikeMirzayanov/testlib` for this toolkit.
* The `cmdline.h` included in this toolkit is the modified version of the code published under the modified BSD license at `https://github.com/tanakh/cmdline` for this toolkit.
* The sample code is released under the MIT license.
